#!/usr/bin/env python
"""
Tigres Split Template example.

This example takes a data files, extracts the data
and performs basic statistics on some of the data (average, median, total) using
a Tigres split template.
"""

import os
import sys
import math

import tigres
import statistics


def extract_data(filename, input_cols, delimiter=",", has_header=True):
    """
    Get the data from the csv file
    :param filename:
    :return: list of lists
    :rtype: list
    """
    if not has_header:
        raise Exception("Header row required!")
    column_headers = None
    data = {}
    tigres.log_debug("extract_data", message="filename:{}, input_cols:{}".format(filename, input_cols))
    with open(filename, 'rU') as myfile:
        for line in myfile:
            values = line.strip("\n").split(delimiter)
            tigres.log_trace("extract_data", message="line: {}".format(values))
            if not column_headers and has_header:
                column_headers = values
            else:
                for idx, val in enumerate(values):
                    if idx in input_cols:
                        try:
                            anumber = float(val)
                            data.setdefault(column_headers[idx], [])
                            data[column_headers[idx]].append(anumber)
                        except Exception as e:
                            tigres.log_debug("extract_data",
                                            message="Exception '"'{}'"' caught on column {} for value {}. Skipping this row".format(e,
                                                idx, val))

    return data


def median(numeric_values):
    """
    Calculate the median of the given numeric values

    :param numeric_values:
    :type numeric_values: list
    :return: median
    :rtype: float
    """
    values = sorted(numeric_values)
    if len(values) % 2 == 1:
        return values[int((len(values) + 1) / 2 - 1)]
    else:
        lower = values[int(len(values) / 2 - 1)]
        upper = values[int(len(values) / 2)]
        return (float(lower + upper)) / 2


def average_by_column(data, column):
    """
    Calculate the average of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column: name of the column
    :type column: str
    :return: (operation, column_name, average)
    :rtype: tuple
    """
    values = []
    total = sum(data)
    count = len(data)
    average = total / count
    return average_by_column.__name__, column, average


def total_by_column(data, column):
    """
    Calculate the total of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column: name of the column
    :type column: str
    :return: (operation, column_name, total)
    :rtype: tuple
    """
    total = sum(data)
    return total_by_column.__name__, column, total


def median_by_column(data, column):
    """
    Calculate the median of the given numeric values in the specified column

    :param data:
    :type data: list
    :param column: name of the column
    :type column: str
    :return: (operation, column_name, median)
    :rtype: tuple
    """

    return median_by_column.__name__, column, median(data)


def stdev_by_column(data, column):
    return stdev_by_column.__name__, column, math.sqrt(statistics.variance(data))


def variance_by_column(data, column):
    return variance_by_column.__name__, column, statistics.variance(data)


def main(execution, data_file, delimiter, columns):
    # Initialize the Tigres program

    tigres.start(name="Basic Statistics By Column", log_dest=os.path.splitext(__file__)[0] + '.log',
                 execution=tigres.utils.Execution.get(execution))

    tigres.set_log_level(tigres.Level.INFO)
    tigres.log_info('parameters', data_file=os.path.split(data_file)[1], execution=execution, columns=columns)

    # Initialize the field values that we are interested in
    input_cols = [int(c) for c in columns.split(',')]

    # Initialize the operations we would like to perform on the data columns
    impls = [average_by_column, total_by_column, median_by_column, stdev_by_column, variance_by_column]

    # Data Extraction Task
    task_data_extraction = tigres.Task("Data Extraction", tigres.FUNCTION, impl_name=extract_data,
                                       input_types=[str, list, str])
    inputs_data_extraction = tigres.InputValues("Data File", [data_file, input_cols, delimiter])

    data = tigres.sequence('Data Extract', tigres.TaskArray(tasks=[task_data_extraction]),
                           tigres.InputArray(values=[inputs_data_extraction]))
    tigres.log_debug("main", message="Extracted Data", output=data)

    # Define Tasks for statistics
    tasks_statistics = []
    inputs_statistics = []
    for col in data.keys():
        for impl in impls:
            # Notice the used of PREVIOUS to get the output from the Data Extraction task.
            inputs_statistics.append(tigres.InputValues('Column', [data[col], col]))
            task_name = "Task.%s.%s" % (col, impl.__name__)
            tasks_statistics.append(
                tigres.Task(task_name, task_type=tigres.FUNCTION, impl_name=impl))

    output = tigres.parallel('Data Stats', tigres.TaskArray(tasks=tasks_statistics),
                             tigres.InputArray(values=inputs_statistics))


    # Print out the output returns by the split template
    tigres.log_info("result_count", message="Parallel Output Result Count {}".format(len(output)))
    for output in output:
        if isinstance(output, tigres.utils.TaskFailure):
            print(output.error)
        else:
            tigres.log_info("results", message=str(output))



if __name__ == "__main__":

    if len(sys.argv) <= 3:
        print("\nBasic Statistics Workflow (mean, median and total)\n")
        print(("Usage: {} <execution plugin> <data file> <delimiter> <columns>\n".format(
            sys.argv[0])))
        print(
        "\texecution plugin - possible values are ({})".format("|".join(list(tigres.utils.Execution.LOOKUP.keys()))))
        print("\tdata file - comma delimited text file with a header row")
        print("\tdelimiter - The file delimiter. Typically comma.")
        print("\tcolumns (Optional) list of column number (1,2,3) or blank for all")
        exit()
    try:
        main(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])

        log_records = tigres.query(["event = results"])
        for record in log_records:
            print("{} - {}".format(record.event, record.message))

    except tigres.utils.TigresException as e:
        print(e)

        # Check the status of the statistical tasks in the log
        task_check = tigres.check('task', state=tigres.utils.State.FAIL)
        if task_check:
            for task_record in task_check:
                print(".. State of {} = {} -  {}".format(task_record.name, task_record.state, task_record.errmsg))

        else:
            print("Success!")

    tigres.dot_execution()

    tigres.end()
